"""Tests for the GenerationStep classes."""

import pytest

from pydexpi.syndata.connector_renaming import ConnectorRenamingConvention
from pydexpi.syndata.generator_step import (
    AddPatternStep,
    CappingStep,
    InitializationStep,
    InternalConnectionStep,
    TerminationStep,
)
from tests.test_syndata.test_pattern import DummyConnector, DummyPattern


def test_add_pattern_step_to_dict():
    """Test that AddPatternStep.to_dict returns correct dictionary representation."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")
    next_pat = DummyPattern("pattern", connectors=[next_conn])

    step = AddPatternStep(own, next_pat, next_conn, "sampled")

    d = step.to_dict()
    assert d["generator_step_type"] == "add_pattern"
    assert d["own_connector"] == "own"
    assert d["next_pattern"] == "pattern"
    assert d["next_connector"] == "next"
    assert d["sampled_distribution_name"] == "sampled"


def test_add_pattern_step_execute():
    """Test that executing AddPatternStep incorporates the next pattern."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")
    next_pat = DummyPattern("pattern", connectors=[next_conn])
    current_pat = DummyPattern("current", connectors=[own])

    # Test with an invalid own connector
    invalid_own = DummyConnector("invalid")
    step = AddPatternStep(invalid_own, next_pat, next_conn, "sampled")
    with pytest.raises(RuntimeError, match="is not associated"):
        step.execute_on(current_pat)

    # Test with a valid own connector
    step = AddPatternStep(own, next_pat, next_conn, "sampled")

    step.execute_on(current_pat)
    assert next_pat.is_incorporated is True


def test_add_pattern_step_invalid_own_connector():
    """Test that executing AddPatternStep with an invalid own_connector raises RuntimeError."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")
    next_pat = DummyPattern("pattern", connectors=[next_conn])
    current_pat = DummyPattern("current", connectors=[own])

    # Test with an invalid own connector
    invalid_own = DummyConnector("invalid")
    step = AddPatternStep(invalid_own, next_pat, next_conn, "sampled")
    with pytest.raises(RuntimeError, match="is not associated"):
        step.execute_on(current_pat)


def test_add_pattern_step_invalid_next_connector():
    """Test that initializing AddPatternStep with an invalid next_connector raises ValueError."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")
    next_pat = DummyPattern("pattern", connectors=[])  # Empty connectors list to trigger error
    with pytest.raises(ValueError, match="The connector"):
        AddPatternStep(own, next_pat, next_conn, "sampled")


def test_add_pattern_step_termination_status():
    """Test that AddPatternStep returns termination status False."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")
    next_pat = DummyPattern("pattern", connectors=[next_conn])

    step = AddPatternStep(own, next_pat, next_conn, "sampled")
    assert step.get_termination_status() is False


def test_add_pattern_step_apply_renaming_convention():
    """Test that AddPatternStep applies the renaming convention."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")
    other_conn = DummyConnector("other")
    next_pat = DummyPattern("pattern", connectors=[next_conn, other_conn])

    step = AddPatternStep(own, next_pat, next_conn, "sampled")
    renaming_convention = ConnectorRenamingConvention()
    step.apply_renaming_convention(renaming_convention=renaming_convention)

    # Own and next stay the same, other gets renamed
    assert own.label == "own"
    assert next_conn.label == "next"
    assert other_conn.label == "pattern_0_other"


def test_interal_connection_step_with_same_connectors_raises_value_error():
    """Test that initializing InternalConnectionStep with the same connector raises ValueError."""
    own = DummyConnector("own")

    with pytest.raises(ValueError, match="The connector"):
        InternalConnectionStep(own, own)


def test_internal_connection_step_to_dict():
    """Test that InternalConnectionStep.to_dict returns correct dictionary representation."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")

    step = InternalConnectionStep(own, next_conn)

    d = step.to_dict()
    assert d["generator_step_type"] == "internal_connection"
    assert d["own_connector"] == "own"
    assert d["next_connector"] == "next"
    assert d["next_pattern"] is None
    assert d["sampled_distribution_name"] is None


def test_internal_connection_step_execute():
    """Test that executing InternalConnectionStep deactivates the own connector."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")
    current_pat = DummyPattern("current", connectors=[own, next_conn])

    # Test with a valid own connector
    step = InternalConnectionStep(own, next_conn)

    step.execute_on(current_pat)
    assert own.is_active is False


def test_internal_connection_step_execute_runtime_error():
    """Test that executing InternalConnectionStep with a missing next_connector raises RuntimeError."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")
    current_pat = DummyPattern("current", connectors=[own, next_conn])

    invalid_conn = DummyConnector("invalid")

    step = InternalConnectionStep(own, invalid_conn)
    with pytest.raises(RuntimeError, match="is not associated"):
        step.execute_on(current_pat)

    step = InternalConnectionStep(invalid_conn, next_conn)
    with pytest.raises(RuntimeError, match="is not associated"):
        step.execute_on(current_pat)


def test_internal_connection_step_termination_status():
    """Test that InternalConnectionStep returns termination status False."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")

    step = InternalConnectionStep(own, next_conn)
    assert step.get_termination_status() is False


def test_termination_step():
    """Test that TerminationStep returns termination status and leaves connectors unchanged."""
    dummy_pat = DummyPattern("dummy", connectors=[DummyConnector("dummy")])
    step = TerminationStep()
    step.execute_on(dummy_pat)
    d = step.to_dict()
    assert d["generator_step_type"] == "termination"
    assert d["own_connector"] is None
    assert d["next_connector"] is None
    assert d["next_pattern"] is None
    assert d["sampled_distribution_name"] is None
    assert step.get_termination_status() is True


def test_initialization_step():
    """Test that InitializationStep returns a proper dictionary and pattern."""
    pat = DummyPattern("init", connectors=[DummyConnector("init")])
    step = InitializationStep(pat, "sampled")
    d = step.to_dict()
    assert d["generator_step_type"] == "initialization"
    assert d["next_pattern"] == "init"
    assert d["sampled_distribution_name"] == "sampled"
    assert step.get_pattern() == pat


def test_capping_step_to_dict():
    """Test that CappingStep.to_dict returns correct dictionary representation."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")
    next_pat = DummyPattern("pattern_cap", connectors=[next_conn])

    step = CappingStep(own, next_pat, next_conn, "cap_sample")
    d = step.to_dict()
    assert d["generator_step_type"] == "capping"
    assert d["own_connector"] == "own"
    assert d["next_pattern"] == "pattern_cap"
    assert d["next_connector"] == "next"
    assert d["sampled_distribution_name"] == "cap_sample"


def test_capping_step_execute_with_next_pattern():
    """Test that CappingStep.execute_on incorporates the next pattern."""
    own = DummyConnector("own")
    next_conn = DummyConnector("next")
    next_pat = DummyPattern("pattern_cap", connectors=[next_conn])
    current_pat = DummyPattern("current_cap", connectors=[own])

    step = CappingStep(own, next_pat, next_conn, "cap_sample")
    step.execute_on(current_pat)

    assert next_pat.is_incorporated is True
    assert next_conn.is_active is False
    assert own.is_active is False


def test_capping_step_execute_drop_connector():
    """Test that CappingStep.execute_on drops the connector when next_pattern is None."""
    own = DummyConnector("own")
    current_pat = DummyPattern("current_drop", connectors=[own])
    step = CappingStep(own, None, None, None)
    step.execute_on(current_pat)

    # Verify that the connector is dropped and removed from the pattern
    assert own.is_active is False
    assert current_pat.connectors == {}


def test_capping_step_invalid_connector_and_multiple_connectors():
    """Test that initializing CappingStep with an invalid next_connector or multiple connectors raises ValueError."""
    own = DummyConnector("own")
    invalid_conn = DummyConnector("invalid")
    # Create a pattern with an empty connectors list to trigger invalid connector error
    next_pat_invalid = DummyPattern("pattern_invalid", connectors=[])

    with pytest.raises(ValueError, match="The connector"):
        CappingStep(own, next_pat_invalid, invalid_conn, "cap_sample")

    # Create a pattern with multiple connectors
    conn1 = DummyConnector("conn1")
    conn2 = DummyConnector("conn2")
    next_pat_multi = DummyPattern("pattern_multi", connectors=[conn1, conn2])
    with pytest.raises(ValueError, match="has more than one connector"):
        CappingStep(own, next_pat_multi, conn1, "cap_sample")



from pydexpi.toolkits import model_toolkit as mt, piping_toolkit as pt


# @pytest.mark.dependency(depends=["test_load_proteus"])
def test_parse_proteus_to_dexpi(loaded_example_dexpi):
    """Parse proteus root tree to dexpi classes."""
    dexpi_model = loaded_example_dexpi
    # Check if total number of objects is still correct (for equipment, piping,
    # and instrumentation)
    assert len(mt.get_all_instances_in_model(dexpi_model)) == 214
    # Check the piping network segments
    for system in dexpi_model.conceptualModel.pipingNetworkSystems:
        for segment in system.segments:
            assert (
                pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID
            )
    # Run some misc tests
    assert len(dexpi_model.conceptualModel.actuatingSystems) == 3
    assert (
        dexpi_model.conceptualModel.taggedPlantItems[0].__class__.__name__ == "PlateHeatExchanger"
    )


def test_parse_generic_attributes(loaded_example_dexpi):
    """Test if DEXPI generic attributes are parsed correctly from proteus."""
    dexpi_model = loaded_example_dexpi
    assert dexpi_model.conceptualModel.taggedPlantItems[0].plateHeight.unit.value == "mm"
    assert float(dexpi_model.conceptualModel.taggedPlantItems[0].plateHeight.value) == 850
    assert (
        dexpi_model.conceptualModel.taggedPlantItems[4]
        .nozzles[0]
        .nodes[0]
        .nominalDiameterRepresentation
        == "DN 80"
    )



import pytest

from pydexpi.dexpi_classes import instrumentation
from pydexpi.toolkits import instrumentation_toolkit as it


def test_add_signal_generating_function_to_instrumentation_function():
    """Test adding a signal generating function to a process instrumentation function."""
    # Create a new instrumentation function
    instr_func = instrumentation.ProcessInstrumentationFunction()
    signal_gen = instrumentation.ProcessSignalGeneratingFunction()
    measuring_line = instrumentation.MeasuringLineFunction()

    # Add the signal generating function to the instrumentation function
    it.add_signal_generating_function_to_instrumentation_function(
        instr_func, signal_gen, measuring_line
    )

    # Check that the function was added correctly
    assert measuring_line in instr_func.signalConveyingFunctions
    assert signal_gen in instr_func.processSignalGeneratingFunctions
    assert measuring_line.source == signal_gen
    assert measuring_line.target == instr_func

    # Test error cases
    # Case 1: Measuring line already exists in the instrumentation function
    with pytest.raises(ValueError):
        it.add_signal_generating_function_to_instrumentation_function(
            instr_func, instrumentation.ProcessSignalGeneratingFunction(), measuring_line
        )

    # Case 2: Signal generating function already exists in the instrumentation function
    new_measuring_line = instrumentation.MeasuringLineFunction()
    with pytest.raises(ValueError):
        it.add_signal_generating_function_to_instrumentation_function(
            instr_func, signal_gen, new_measuring_line
        )

    # Case 3: Measuring line already has a source that is not the signal generating function
    new_measuring_line = instrumentation.MeasuringLineFunction()
    new_signal_gen = instrumentation.ProcessSignalGeneratingFunction()
    new_measuring_line.source = new_signal_gen
    with pytest.raises(ValueError):
        it.add_signal_generating_function_to_instrumentation_function(
            instr_func, instrumentation.ProcessSignalGeneratingFunction(), new_measuring_line
        )

    # Case 4: Measuring line already has a target that is not the instrumentation function
    new_measuring_line = instrumentation.MeasuringLineFunction()
    new_instr_func = instrumentation.ProcessInstrumentationFunction()
    new_measuring_line.target = new_instr_func
    with pytest.raises(ValueError):
        it.add_signal_generating_function_to_instrumentation_function(
            instr_func, instrumentation.ProcessSignalGeneratingFunction(), new_measuring_line
        )


def test_add_signal_opc_to_instrumentation_function():
    """Test adding a signal off-page connector to an instrumentation function."""
    # Create a new instrumentation function
    instr_func = instrumentation.ProcessInstrumentationFunction()

    # Test outgoing signal OPC
    outgoing_opc = instrumentation.FlowOutSignalOffPageConnector()
    outgoing_line = instrumentation.SignalLineFunction()

    it.add_signal_opc_to_instrumentation_function(instr_func, outgoing_opc, outgoing_line)

    # Check that the function was added correctly for outgoing signal
    assert outgoing_line in instr_func.signalConveyingFunctions
    assert outgoing_opc in instr_func.signalConnectors
    assert outgoing_line.source == instr_func
    assert outgoing_line.target == outgoing_opc

    # Test incoming signal OPC
    incoming_opc = instrumentation.FlowInSignalOffPageConnector()  # base class is used for incoming
    incoming_line = instrumentation.SignalLineFunction()

    it.add_signal_opc_to_instrumentation_function(instr_func, incoming_opc, incoming_line)

    # Check that the function was added correctly for incoming signal
    assert incoming_line in instr_func.signalConveyingFunctions
    assert incoming_opc in instr_func.signalConnectors
    assert incoming_line.source == incoming_opc
    assert incoming_line.target == instr_func

    # Test error cases
    # Case 1: Signal OPC already exists in the instrumentation function
    with pytest.raises(ValueError):
        it.add_signal_opc_to_instrumentation_function(
            instr_func, outgoing_opc, instrumentation.SignalLineFunction()
        )

    # Case 2: Signal line already exists in the instrumentation function
    with pytest.raises(ValueError):
        it.add_signal_opc_to_instrumentation_function(
            instr_func, instrumentation.FlowOutSignalOffPageConnector(), outgoing_line
        )

    # Case 3: Signal line already has a target (for outgoing signals)
    new_outgoing_line = instrumentation.SignalLineFunction()
    new_outgoing_line.target = instrumentation.FlowOutSignalOffPageConnector()
    with pytest.raises(ValueError):
        it.add_signal_opc_to_instrumentation_function(
            instr_func, instrumentation.FlowOutSignalOffPageConnector(), new_outgoing_line
        )

    # Case 4: Signal line already has a source (for incoming signals)
    new_incoming_line = instrumentation.SignalLineFunction()
    new_incoming_line.source = instrumentation.FlowInSignalOffPageConnector()
    with pytest.raises(ValueError):
        it.add_signal_opc_to_instrumentation_function(
            instr_func, instrumentation.FlowInSignalOffPageConnector(), new_incoming_line
        )


def test_add_actuating_function_to_instrumentation_function():
    """Test adding an actuating function to an instrumentation function."""
    # Create a new instrumentation function
    instr_func = instrumentation.ProcessInstrumentationFunction()

    # Test regular actuating function
    actuating_func = instrumentation.ActuatingFunction()
    signal_line = instrumentation.SignalLineFunction()

    it.add_actuating_function_to_instrumentation_function(instr_func, actuating_func, signal_line)

    # Check that the function was added correctly
    assert signal_line in instr_func.signalConveyingFunctions
    assert actuating_func in instr_func.actuatingFunctions
    assert signal_line.source == instr_func
    assert signal_line.target == actuating_func

    # Test electrical actuating function
    elect_actuating_func = instrumentation.ActuatingElectricalFunction()
    elect_signal_line = instrumentation.SignalLineFunction()

    it.add_actuating_function_to_instrumentation_function(
        instr_func, elect_actuating_func, elect_signal_line
    )

    # Check that the electrical function was added correctly
    assert elect_signal_line in instr_func.signalConveyingFunctions
    assert elect_actuating_func in instr_func.actuatingElectricalFunctions
    assert elect_signal_line.source == instr_func
    assert elect_signal_line.target == elect_actuating_func

    # Test error cases
    # Case 1: Signal line already exists in the instrumentation function
    with pytest.raises(ValueError):
        it.add_actuating_function_to_instrumentation_function(
            instr_func, instrumentation.ActuatingFunction(), signal_line
        )

    # Case 2: Actuating function already exists in the instrumentation function
    new_signal_line = instrumentation.SignalLineFunction()
    with pytest.raises(ValueError):
        it.add_actuating_function_to_instrumentation_function(
            instr_func, actuating_func, new_signal_line
        )

    # Case 3: Electrical actuating function already exists in the instrumentation function
    new_elect_signal_line = instrumentation.SignalLineFunction()
    with pytest.raises(ValueError):
        it.add_actuating_function_to_instrumentation_function(
            instr_func, elect_actuating_func, new_elect_signal_line
        )

    # Case 4: Signal line already has a source that is not the instrumentation function
    new_signal_line = instrumentation.SignalLineFunction()
    new_signal_line.source = instrumentation.ProcessInstrumentationFunction()
    with pytest.raises(ValueError):
        it.add_actuating_function_to_instrumentation_function(
            instr_func, instrumentation.ActuatingFunction(), new_signal_line
        )

    # Case 5: Signal line already has a target that is not the actuating function
    new_signal_line = instrumentation.SignalLineFunction()
    new_signal_line.target = instrumentation.ActuatingFunction()
    with pytest.raises(ValueError):
        it.add_actuating_function_to_instrumentation_function(
            instr_func, instrumentation.ActuatingFunction(), new_signal_line
        )


def test_connect_instrumentation_functions():
    """Test connecting two instrumentation functions via a signal line."""
    # Create two instrumentation functions
    source_func = instrumentation.ProcessInstrumentationFunction()
    target_func = instrumentation.ProcessInstrumentationFunction()
    signal_line = instrumentation.SignalLineFunction()

    # Connect the functions with source_signal_line=True
    it.connect_instrumentation_functions(
        source_func, target_func, signal_line, source_signal_line=True
    )

    # Check that the functions were connected correctly
    assert signal_line in source_func.signalConveyingFunctions
    assert signal_line not in target_func.signalConveyingFunctions
    assert signal_line.source == source_func
    assert signal_line.target == target_func

    # Connect two other functions with source_signal_line=False
    source_func2 = instrumentation.ProcessInstrumentationFunction()
    target_func2 = instrumentation.ProcessInstrumentationFunction()
    signal_line2 = instrumentation.SignalLineFunction()

    it.connect_instrumentation_functions(
        source_func2, target_func2, signal_line2, source_signal_line=False
    )

    # Check that the functions were connected correctly
    assert signal_line2 not in source_func2.signalConveyingFunctions
    assert signal_line2 in target_func2.signalConveyingFunctions
    assert signal_line2.source == source_func2
    assert signal_line2.target == target_func2

    # Test error cases
    # Case 1: Signal line already exists in the source function
    new_signal_line = instrumentation.SignalLineFunction()
    source_func3 = instrumentation.ProcessInstrumentationFunction()
    target_func3 = instrumentation.ProcessInstrumentationFunction()
    source_func3.signalConveyingFunctions.append(new_signal_line)

    with pytest.raises(ValueError):
        it.connect_instrumentation_functions(source_func3, target_func3, new_signal_line)

    # Case 2: Signal line already exists in the target function
    new_signal_line = instrumentation.SignalLineFunction()
    source_func4 = instrumentation.ProcessInstrumentationFunction()
    target_func4 = instrumentation.ProcessInstrumentationFunction()
    target_func4.signalConveyingFunctions.append(new_signal_line)

    with pytest.raises(ValueError):
        it.connect_instrumentation_functions(source_func4, target_func4, new_signal_line)

    # Case 3: Signal line already has a source that is not the source function
    new_signal_line = instrumentation.SignalLineFunction()
    new_signal_line.source = instrumentation.ProcessInstrumentationFunction()

    with pytest.raises(ValueError):
        it.connect_instrumentation_functions(source_func, target_func, new_signal_line)

    # Case 4: Signal line already has a target that is not the target function
    new_signal_line = instrumentation.SignalLineFunction()
    new_signal_line.target = instrumentation.ProcessInstrumentationFunction()

    with pytest.raises(ValueError):
        it.connect_instrumentation_functions(source_func, target_func, new_signal_line)



"""Tests for the module dexpi_pattern."""

import pytest

from pydexpi.dexpi_classes import equipment, piping
from pydexpi.syndata.dexpi_pattern import (
    BasicPipingInConnector,
    BasicPipingOutConnector,
    DexpiPattern,
)
from pydexpi.toolkits import piping_toolkit as pt


@pytest.fixture
def in_connector_factory():
    """Fixture to create an in connector factory."""

    def _create_in_connector():
        """Create an in connector."""
        in_item = piping.OperatedValve(nodes=[piping.PipingNode() for i in range(2)])
        return BasicPipingInConnector("Main-In", in_item, target_node_index=0)

    return _create_in_connector


@pytest.fixture
def out_connector_factory(simple_pns_factory):
    """Fixture to create an out connector factory."""

    def _create_out_connector():
        """Create an out connector."""
        segment = simple_pns_factory()
        pt.append_connection_to_unconnected_segment(segment, piping.Pipe(), 1)
        return BasicPipingOutConnector("Main-Out", segment)

    return _create_out_connector


@pytest.fixture
def the_pattern_factory(simple_dexpi_model_factory, connector_label_prefix="Main"):
    """Fixture to create a pattern factory."""

    def _create_pattern(pattern_name="the_pattern", connector_prefix=connector_label_prefix):
        """Create a Dexpi pattern."""
        model = simple_dexpi_model_factory()
        new_nozzle = equipment.Nozzle(nodes=[piping.PipingNode()])
        model.conceptualModel.taggedPlantItems[0].nozzles.append(new_nozzle)
        in_conn = BasicPipingInConnector(f"{connector_prefix}-In", new_nozzle, target_node_index=0)
        out_conn = BasicPipingOutConnector(
            f"{connector_prefix}-Out",
            model.conceptualModel.pipingNetworkSystems[-1].segments[-1],
        )
        return DexpiPattern(pattern_name, [in_conn, out_conn], model)

    return _create_pattern


# Connector tests
def test_connector_constructor(in_connector_factory, out_connector_factory):
    """Evaluation if the In and Out Connectors are created properly"""
    conn_in = in_connector_factory()
    conn_out = out_connector_factory()
    assert conn_in is not None
    assert conn_out is not None


def test_assert_valid_counterpart(in_connector_factory, out_connector_factory):
    """Test to see if dexpi connectors correctly assert counterpart validity."""
    instance_in = in_connector_factory()
    instance_out = out_connector_factory()
    assert instance_in.assess_valid_counterpart(instance_out)
    assert instance_out.assess_valid_counterpart(instance_in)
    second_in = in_connector_factory()
    second_out = out_connector_factory()
    assert not instance_in.assess_valid_counterpart(second_in)
    assert not instance_out.assess_valid_counterpart(second_out)


def test_connect_to_counterpart(in_connector_factory, out_connector_factory):
    """Try connecting the connectors. Call the method on the in connector,
    because this one calls the connection method of its counterpart."""
    instance_in = in_connector_factory()
    instance_out = out_connector_factory()
    second_in = in_connector_factory()
    with pytest.raises(ValueError):
        instance_in.connect_to_counterpart(second_in)

    instance_in.connect_to_counterpart(instance_out)
    assert instance_out.piping_network_segment.targetItem == instance_in.target_item
    assert instance_out.piping_network_segment.targetNode == instance_in.target_item.nodes[0]


# Test pattern functionality
def test_pattern_constructor(the_pattern_factory):
    """Evaluation if the pattern constructor works as required."""
    pattern_instance = the_pattern_factory()
    assert len(pattern_instance.connectors) == 2


def test_incorporate_pattern(the_pattern_factory):
    """Test incorporating a counterpart to a pattern to a counterpart."""

    new_pattern = the_pattern_factory("Pattern 1")
    counterpart = the_pattern_factory("Pattern 2", connector_prefix="Counterpart")

    new_pattern.incorporate_pattern(
        new_pattern.connectors["Main-Out"], counterpart, counterpart.connectors["Counterpart-In"]
    )
    assert len(new_pattern.connectors) == 2
    assert len(new_pattern.dexpi_model.conceptualModel.taggedPlantItems) == 2

    # Check if exceptions are working if connectors are not correctly affiliated
    new_pattern = the_pattern_factory()
    counterpart = the_pattern_factory()
    # Removing the out connector from pattern by key
    faulty_connector = new_pattern.connectors["Main-Out"]
    new_pattern.connectors.pop("Main-Out")
    with pytest.raises(ValueError):
        new_pattern.incorporate_pattern(
            faulty_connector, counterpart, counterpart.connectors["Main-In"]
        )

    new_pattern = the_pattern_factory()
    counterpart = the_pattern_factory()
    # Removing the in connector from the counterpart by key
    faulty_connector = counterpart.connectors["Main-In"]
    counterpart.connectors.pop("Main-In")
    with pytest.raises(ValueError):
        new_pattern.incorporate_pattern(
            new_pattern.connectors["Main-Out"], counterpart, faulty_connector
        )


def test_load_and_save(tmp_path, the_pattern_factory):
    """Test loading and saving a graph pattern"""
    pattern_instance = the_pattern_factory()
    pattern_instance.save(tmp_path, "test")
    loaded_pattern = DexpiPattern.load(tmp_path, "test")

    assert pattern_instance.label == loaded_pattern.label

    connectors = pattern_instance.connectors
    loaded_connectors = loaded_pattern.connectors
    assert len(connectors) == len(loaded_connectors)
    for c, lc in zip(connectors, loaded_connectors):
        assert c == lc
        assert connectors[c].label == loaded_connectors[lc].label

    observers = pattern_instance.observer_patterns
    loaded_observers = loaded_pattern.observer_patterns
    assert len(observers) == len(loaded_observers)
    assert set(observers.keys()) == set(loaded_observers.keys())



"""Core module for the framework of loading the Proteus XML format.

The proteus loading is built as a modular framework. This allows for easy
extension and customization of the loading process. This may be necessary for differently structured
or incorrectly structured Proteus XML files.

The core component of the framework is the `LoaderModule` class. This abstract base class defines
the module interface for loading specific XML elements. Each module is responsible for
handling a specific XML element type and how the data of its contained modules is integrated.
Note that the loader module is responsible for calling the compositional and reference passes of its
submodules at the appropriate time. All created objects are registered in the central
`ObjectRegistry` for later reference resolution, and errors are collected in the central
`ErrorRegistry`.

The loading process consists of three main passes:
1. **Compositional Pass**: creates the object and all compositional relationships between them.
2. **Reference Pass**: Resolves all remaining references between objects. This is done after
   all objects have been created, allowing for a complete view of the object graph.
3. **Control Pass**: Checks the consistency of duplicate information, e.g. object associations.

To facilitate modularity, the loader modules are instantiated by a `ParserFactory`, which is
responsible for creating the appropriate loader module based on the XML element type and plugging
it into parent parsers during construction.

Loading is orchestrated by the `ProteusLoader` class, which uses the `ParserFactory` to create the
top-level model parser for a Proteus XML. It can be used to load XML strings or files.
"""

import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Protocol

from pydexpi.dexpi_classes.pydantic_classes import DexpiBaseModel, DexpiModel


class InternalParserError(Exception):
    """Custom exception for internal parser errors.

    This exception is raised when an error occurs during the parsing process that is not
    directly related to the XML structure or data, but rather to the internal logic of the parser.

    If this exception is raised, it indicates a bug in the parser code that needs to be fixed, and
    should therefore not be caught by the parser.
    """


class ErrorLevels(Enum):
    """Enumeration of error severity levels for the Proteus serializer.

    This enum defines the different levels of errors that can occur
    during the parsing and serialization process.

    Levels:
    - INFO: Informational messages. Non-critical for logging.
    - WARNING: Indicates a potential issue or inconsistency, but no loss of DEXPI data.
    - ERROR: Indicates an error that causes a local loss of DEXPI data, due to missing data or
             malformed XML element.
    - CRITICAL: Indicates a severe error that prevents loading the model entirely.
    """

    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    CRITICAL = "critical"


@dataclass
class ErrorRegistryEntry:
    """Data class representing an error entry in the error registry.

    Parameters
    ----------
    message : str
        Human-readable error message
    level : ErrorLevels
        Severity level of the error
    proteus_id : str or None, optional
        Proteus element ID associated with the error if available, by default None
    exception : Exception or None, optional
        The exception that caused the error, if applicable, by default None
    """

    message: str
    level: ErrorLevels
    proteus_id: str | None = None
    exception: Exception | None = None


class ErrorRegistry:
    """Registry for collecting and managing errors during serialization.

    This class provides a centralized way to collect, store, and retrieve
    errors that occur during the parsing and serialization process.
    """

    def __init__(self) -> None:
        """Initialize an empty error registry."""
        self.errors = []

    def register_error(self, error: ErrorRegistryEntry) -> None:
        """Register a new error in the registry.

        This method is enables centralized error registration. It keeps track of all errors and
        centrally handles InternalParserErrors, which are automatically reraised.

        Parameters
        ----------
        error : ErrorRegistryEntry
            The error entry to register
        """
        self.errors.append(error)
        if isinstance(error.exception, InternalParserError):
            # If the error is an InternalParserError, it should not be caught by the parser.
            raise error.exception

    def get_errors(self, levels: list[ErrorLevels] = None) -> list[ErrorRegistryEntry]:
        """Retrieve all registered errors.

        Parameters
        ----------
        levels : list[ErrorLevels] or None, optional
            If provided, filter errors by their severity level. If None, return all errors.
            Defaults to None.

        Returns
        -------
        list[ErrorRegistryEntry]
            List of all error entries in the registry
        """
        if levels is None:
            return self.errors
        else:
            return [error for error in self.errors if error.level in levels]


class ObjectRegistry:
    """Registry for managing objects created during serialization.

    This class provides a way to store and retrieve objects by their Proteus IDs, enabling reference
    resolution during deserialization.
    """

    def __init__(self) -> None:
        """Initialize an empty object registry."""
        self.objects = {}

    def register_object(self, proteus_id: str, obj: Any) -> None:
        """Register an object in the registry with its Proteus ID.

        Parameters
        ----------
        proteus_id : str
            The Proteus ID to associate with the object
        obj : Any
            The object to register
        """
        self.objects[proteus_id] = obj

    def get_object(self, proteus_id: str) -> Any | None:
        """Retrieve an object by its Proteus ID.

        Parameters
        ----------
        proteus_id : str
            The Proteus ID of the object to retrieve

        Returns
        -------
        Any
            The object associated with the ID, or None if not found
        """
        return self.objects.get(proteus_id)


@dataclass
class ModuleContext:
    """Data class representing the context for a module.

    This class encapsulates the necessary context information for a loader module,
    including the element stack, factory dispatcher, error registry, and object registry.

    Parameters
    ----------
    element_stack : list[str]
        Stack of XML element tags representing the current parsing context.
    id_stack : list[str]
        Stack of Proteus IDs for the current parsing context.
    error_registry : ErrorRegistry
        Registry for collecting errors during loading
    object_registry : ObjectRegistry
        Registry for managing created objects during loading
    """

    element_stack: list[str]
    id_stack: list[str]
    error_registry: ErrorRegistry
    object_registry: ObjectRegistry

    def get_updated_context(self, element: ET.Element) -> "ModuleContext":
        """Create a new ModuleContext with the current element and ID appended to the stacks.

        Parameters
        ----------
        element : ET.Element
            The XML element being processed

        Returns
        -------
        ModuleContext
            A new ModuleContext instance with updated element and ID stacks.
        """
        return ModuleContext(
            element_stack=self.element_stack + [element.tag],
            id_stack=self.id_stack + [element.get("ID", None)],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )

    def get_last_id(self) -> str | None:
        """Get the last ID in the ID stack that isn't None.

        Returns
        -------
        str | None
            The last ID in the stack, or None if the stack is empty.
        """
        for id in reversed(self.id_stack):
            if id is not None:
                return id
        raise ValueError("ID stack is empty or contains only None values.")

    def register_error(
        self, message: str, level: ErrorLevels, proteus_id: str = None, exception: Exception = None
    ) -> None:
        """Register an error with the error registry.

        Parameters
        ----------
        message : str
            Human-readable error message describing the issue
        level : ErrorLevels
            The severity level of the error
        proteus_id : str
            Optional ID associated with the Proteus element that caused the error. If not provided,
            the last ID from the context's ID stack is used.
        exception : Exception
            Optional exception instance related to the error
        """
        if proteus_id is None:
            try:
                proteus_id = self.get_last_id()
            except ValueError:
                proteus_id = None
        new_error = ErrorRegistryEntry(
            message=message,
            level=level,
            proteus_id=proteus_id,
            exception=exception,
        )
        self.error_registry.register_error(new_error)


class ParserModule(ABC):
    """Abstract base class for XML element loader modules.

    This class defines the interface for modules that handle the loading and processing of specific
    XML elements during deserialization. Also includes some utilities for handling the context.

    Attributes
    ----------
    context : ModuleContext
        The context containing the ID-, and element stack, error-, and object registry.
    """

    def __init__(
        self,
        context: ModuleContext,
    ) -> None:
        """Initialize a loader module.

        Parameters
        ----------
        context : ModuleContext
            The context containing the ID-, and element stack, error-, and object registry.
        """
        self.context = context
        self.submodules: list[ParserModule] = []

    @abstractmethod
    def compositional_pass(self) -> Any:
        """Perform the compositional pass of loading.

        This method handles the creation of objects and their compositional relationships during the
        first pass of deserialization.

        Returns
        -------
        Any
            The created object or result of the compositional pass
        """

    def reference_pass(self) -> None:
        """Perform the reference pass of loading.

        This method handles the resolution of object references during the second pass of
        deserialization. Implementation optional, as some modules may not require a reference pass.

        By default, calls the `reference_pass` method of all submodules.
        """
        for submodule in self.submodules:
            submodule.reference_pass()

    def control_pass(self) -> None:
        """This, final pass of the loading process checks the consistency of duplicate info.

        Some information is stored redundantly in the XML file, e.g. object associations. This pass
        checks that the duplicate information is consistent/tries to use the reduntant information
        to fill in missing information.

        By default, this method calls the `control_pass` method of all submodules.
        """
        for submodule in self.submodules:
            submodule.control_pass()

    def register_submodule(self, submodule: "ParserModule") -> None:
        """Register a submodule for the current module.

        This method allows the module to keep track of its submodules, which can be used
        during the compositional, reference, and control passes.

        Parameters
        ----------
        submodule : ParserModule
            The submodule to register
        """
        self.submodules.append(submodule)

    def register_submodule_list(self, submodules: list["ParserModule"]) -> None:
        """Register a list of submodules for the current module.

        This method allows the module to register multiple submodules at once.

        Parameters
        ----------
        submodules : list[ParserModule]
            The list of submodules to register
        """
        for submodule in submodules:
            self.register_submodule(submodule)

    def register_error(
        self, message: str, level: ErrorLevels, proteus_id: str = None, exception: Exception = None
    ) -> None:
        """Register an error with the error registry.

        This method allows subclasses to register errors that occur during
        the loading process. It should be called whenever an error is encountered.

        Parameters
        ----------
        message : str
            Human-readable error message describing the issue
        level : ErrorLevels
            The severity level of the error
        proteus_id : str
            Optional ID associated with the Proteus element that caused the error. If not provided,
            the last ID from the context's ID stack is used.
        exception : Exception
            Optional exception instance related to the error
        """
        self.context.register_error(
            message=message,
            level=level,
            proteus_id=proteus_id,
            exception=exception,
        )

    def register_object(self, proteus_id: str, obj: DexpiBaseModel) -> None:
        """Register an object with the object registry.

        This method allows subclasses to register objects that are created during the loading
        process. It should be called whenever a new object is instantiated. That way, the object can
        be referenced later during the reference pass.

        Parameters
        ----------
        proteus_id : str
            The Proteus ID to associate with the object, used as key for later retrieval.
        obj : Any
            The object to register

        """
        self.context.object_registry.register_object(proteus_id=proteus_id, obj=obj)

    def get_object_from_registry(self, proteus_id: str) -> DexpiBaseModel | None:
        """Retrieve an object from the object registry by its Proteus ID.

        This method allows subclasses to access objects that have been registered
        during the loading process.

        Parameters
        ----------
        proteus_id : str
            The Proteus ID of the object to retrieve

        Returns
        -------
        DexpiBaseModel | None
            The object associated with the ID, or None if not found
        """
        return self.context.object_registry.get_object(proteus_id)


class PlantModelParserProtocol(Protocol):
    """Protocol for a parser that handles the entire plant model.

    This protocol defines the interface for a parser that processes the top-level plant model
    in a Proteus XML file. It includes methods for the compositional, reference, and control passes.
    """

    def compositional_pass(self) -> DexpiModel:
        """Perform the compositional pass of loading the plant model.

        Returns
        -------
        DexpiModel
            The created DEXPI model object
        """

    def reference_pass(self) -> None:
        """Perform the reference pass of loading the plant model to resolve references."""

    def control_pass(self) -> None:
        """Perform the control pass of loading the plant model to validate consistency."""


class ParserFactoryProtocol(Protocol):
    """Protocol for a factory that creates parser modules for Proteus XML elements.

    This protocol defines the interface for creating parser modules based on the XML element type.
    It allows for flexible and modular parsing of different Proteus XML structures.
    """

    def make_plant_model_parser(
        self, context: ModuleContext, element: ET.Element
    ) -> PlantModelParserProtocol:
        """Create a PlantModelParser for the given XML element and context.

        Parameters
        ----------
        context : ModuleContext
            The context in which the parser operates
        element : ET.Element
            The XML element to parse


        Returns
        -------
        PlantModelParser
            An instance of PlantModelParser for the given element and context
        """


class ProteusLoader:
    """Main loader class for processing Proteus XML files.

    This class coordinates the loading process by calling the parser factory to create a top-level
    model parser
    """

    def __init__(self, parser_factory: ParserFactoryProtocol) -> None:
        """Initialize the Proteus loader.

        Parameters
        ----------
        parser_factory : ParserFactoryProtocol
            Factory for creating the set-up plant model parser
        """
        self.parser_factory = parser_factory
        self.error_registry = None
        self.object_registry = None

    def load_xmlstring(self, xml_string: str) -> DexpiModel:
        """Load and process an XML string.

        This method performs a two-pass loading process:
        1. Compositional pass: Creates objects and their composition relationships
        2. Reference pass: Resolves object references
        3. Control pass: Validates and finalizes the model

        Parameters
        ----------
        xml_string : str
            The XML content to parse and load

        Returns
        -------
        DexpiModel
            The loaded DEXPI model object
        """
        # Make a new error registry and object registry
        self.reset_registries()

        # Make a root context element for the xml
        root_element = ET.fromstring(xml_string)
        root_id = None
        root_context = ModuleContext(
            element_stack=[root_element.tag],
            id_stack=[root_id],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )
        plant_model_parser = self.parser_factory.make_plant_model_parser(root_context, root_element)
        depxi_model = plant_model_parser.compositional_pass()
        plant_model_parser.reference_pass()
        plant_model_parser.control_pass()

        return depxi_model

    def load_xml_file(self, xml_file: Path) -> Any:
        """Load and process an XML file.

        This method performs a two-pass loading process:
        1. Compositional pass: Creates objects and their composition relationships
        2. Reference pass: Resolves object references

        Parameters
        ----------
        xml_file : Path
            The path to the XML file to parse and load

        Returns
        -------
        Any
            The loaded DEXPI model object
        """
        with open(xml_file) as file:
            xml_string = file.read()

        return self.load_xmlstring(xml_string)

    def get_error_registry(self) -> ErrorRegistry:
        """Get the error registry containing any errors from the loading process.

        Returns
        -------
        ErrorRegistry
            The error registry with collected errors
        """
        return self.error_registry

    def reset_registries(self) -> None:
        """Reset the error and object registries to empty instances for a new loading process."""
        self.error_registry = ErrorRegistry()
        self.object_registry = ObjectRegistry()



"""Tests for the module pattern."""

import pytest

from tests.test_syndata.conftest import DummyConnector, DummyPattern


# Connector tests
def test_connector_constructor():
    """Test constructing a connector."""
    the_connector = DummyConnector("Testlabel", kwinfos={"dummy_kw": "Dummy kw value"})
    # Check if attributes are assigned correctly
    assert the_connector.label == "Testlabel"
    assert "dummy_kw", "Dummy kw value" in the_connector.kwinfos.items()


def test_assess_valid_counterpart():
    """Quick test to check connector method assess_valid_counterpart."""
    the_connector = DummyConnector("Testlabel")
    other_connector = DummyConnector("Testlabel")
    assert the_connector.assess_valid_counterpart(other_connector)


def test_connect_to_counterpart():
    """Test connecting a connector to a counterpart with the
    connect_to_counterpart method."""
    the_connector = DummyConnector("Testlabel")

    # Try to connect a connector to itself:
    with pytest.raises(ValueError):
        the_connector.connect_to_counterpart(the_connector)

    # Connect a counterpart
    the_counterpart = DummyConnector("Counterpart")
    the_connector.connect_to_counterpart(the_counterpart)
    assert the_connector.is_active is False
    assert the_counterpart.is_active is False

    # Try to connect again, and get a runtime error
    new_counterpart = DummyConnector("Counterpart")
    with pytest.raises(RuntimeError):
        the_connector.connect_to_counterpart(new_counterpart)


def test_deactivate():
    """Test deactivating a connector with the deactivate method."""
    the_connector = DummyConnector("Testlabel")

    # Test deactivation
    the_connector.deactivate()

    assert not the_connector.is_active

    # Try to deactivate the connector again:
    with pytest.raises(RuntimeError):
        the_connector.deactivate()

    # Try to connect the deactivated connector
    the_counterpart = DummyConnector("Counterpart")
    with pytest.raises(RuntimeError):
        the_connector.connect_to_counterpart(the_counterpart)


# Pattern tests
def test_pattern_constructor():
    """Test instantiating a simple pattern."""
    the_connector = DummyConnector("Connector label")
    the_pattern = DummyPattern(
        "Pattern label",
        connectors=[the_connector],
        kwinfos={"dummy_kw": "Dummy kw value"},
    )
    assert the_pattern.label == "Pattern label"
    assert the_connector in the_pattern.connectors.values()
    assert "dummy_kw", "Dummy kw value" in the_pattern.kwinfos.items()


def test_add_observer():
    """Test adding observer patterns to a pattern."""
    the_connector = DummyConnector("Connector label")
    the_pattern = DummyPattern("Pattern label", connectors=[the_connector])

    # Try adding an observer with an invalid tag
    observer_connector = DummyConnector("Connector label")
    observer_pattern = DummyPattern("Invalid pattern label", connectors=[observer_connector])
    with pytest.raises(ValueError):
        the_pattern.add_observer("Observer tag", observer_pattern)

    # Try adding an observer with an invalid connector
    observer_connector = DummyConnector("Invalid connector label")
    observer_pattern = DummyPattern("Pattern label", connectors=[observer_connector])
    with pytest.raises(ValueError):
        the_pattern.add_observer("Observer tag", observer_pattern)

    # Try adding an observer with an different number of connectors
    observer_connector = DummyConnector("Connector label")
    observer_connector2 = DummyConnector("Connector2")
    observer_pattern = DummyPattern(
        "Pattern label", connectors=[observer_connector, observer_connector2]
    )
    with pytest.raises(ValueError):
        the_pattern.add_observer("Observer tag", observer_pattern)

    # Add an observer normally
    observer_connector = DummyConnector("Connector label")
    observer_pattern = DummyPattern("Pattern label", connectors=[observer_connector])
    the_pattern.add_observer("Observer tag", observer_pattern)
    assert "Observer tag", observer_pattern in the_pattern.observer_patterns.items()


def test_incorporate_pattern(simple_pattern_factory):
    pattern1 = simple_pattern_factory("Pattern1")
    pattern2 = simple_pattern_factory(
        "Pattern2", connector_label_prefix="Pattern2 Connector label_"
    )

    invalid_connector = DummyConnector("Some label")
    with pytest.raises(ValueError):
        pattern1.incorporate_pattern(
            invalid_connector, pattern2, list(pattern2.connectors.values())[0]
        )
    with pytest.raises(ValueError):
        pattern1.incorporate_pattern(
            list(pattern1.connectors.values())[0], pattern2, invalid_connector
        )

    new_observer = simple_pattern_factory(
        "Pattern2", connector_label_prefix="Pattern2 Connector label_"
    )
    pattern2.add_observer("Observer_tag", new_observer)
    with pytest.raises(ValueError):
        pattern1.incorporate_pattern(
            list(pattern1.connectors.values())[0], pattern2, list(pattern2.connectors.values())[1]
        )

    pattern2 = simple_pattern_factory(
        "Pattern3", connector_label_prefix="Pattern3 Connector label_"
    )
    pattern1.incorporate_pattern(
        list(pattern1.connectors.values())[0], pattern2, list(pattern2.connectors.values())[1]
    )
    assert len(pattern1.connectors) == 2
    assert list(pattern2.connectors.keys())[0] in pattern1.connectors
    assert len(pattern1.observer_patterns["Observer tag"].connectors) == 2
    assert (
        list(pattern2.observer_patterns["Observer tag"].connectors.keys())[0]
        in pattern1.observer_patterns["Observer tag"].connectors
    )
    assert pattern2.is_incorporated is True
    assert pattern1.is_incorporated is False

    new_observer = simple_pattern_factory(
        "Pattern2", connector_label_prefix="Pattern2 Connector label_"
    )
    with pytest.raises(RuntimeError):
        pattern2.add_observer("Observer_tag", new_observer)
    pattern3 = simple_pattern_factory(
        "Pattern3", connector_label_prefix="Pattern3 Connector label_"
    )
    with pytest.raises(RuntimeError):
        pattern2.incorporate_pattern(
            list(pattern2.connectors.keys())[0], pattern3, list(pattern3.connectors.keys())[0]
        )


def test_connect_internal(simple_pattern_factory):
    pattern1 = simple_pattern_factory("Pattern1")
    invalid_connector = DummyConnector("Some label")
    connector1 = list(pattern1.connectors.values())[0]
    connector2 = list(pattern1.connectors.values())[1]
    with pytest.raises(ValueError):
        pattern1.connect_internal(connector1, invalid_connector)
    pattern1.connect_internal(connector1, connector2)
    assert len(pattern1.connectors) == 0
    assert len(pattern1.observer_patterns["Observer tag"].connectors) == 0
    assert connector1.is_active is False
    assert connector2.is_active is False


def test_drop_connector(simple_pattern_factory):
    pattern1 = simple_pattern_factory("Pattern1")
    the_connector = list(pattern1.connectors.values())[0]
    no_connectors = len(pattern1.connectors)
    pattern1.drop_connector(the_connector)
    assert not the_connector._is_active
    assert the_connector not in pattern1.connectors
    assert len(pattern1.connectors) == no_connectors - 1
    assert len(pattern1.observer_patterns["Observer tag"].connectors) == no_connectors - 1


def test_relabel_connector(simple_pattern_factory):
    """Test relabeling a connector with the relabel_connector method."""
    the_pattern = simple_pattern_factory("Pattern1")

    # Get the first connector and its current label
    the_connector = list(the_pattern.connectors.values())[0]
    old_label = the_connector.label

    # Define a new label and relabel the connector
    new_label = "New Connector Label"
    the_pattern.relabel_connector(the_connector, new_label)

    # Check that the connector's label has been updated
    assert the_connector.label == new_label

    # Check that the connectors dictionary has been updated correctly
    assert new_label in the_pattern.connectors
    assert old_label not in the_pattern.connectors
    assert the_pattern.connectors[new_label] == the_connector

    # Check that observer patterns' connectors have also been updated
    observer = the_pattern.observer_patterns["Observer tag"]
    assert new_label in observer.connectors
    assert old_label not in observer.connectors
    assert observer.connectors[new_label].label == new_label


def test_copy_pattern(simple_pattern_factory):
    the_pattern = simple_pattern_factory("Pattern1")
    copied_pattern = the_pattern.copy_pattern()
    assert len(copied_pattern.connectors) == len(the_pattern.connectors)
    assert len(copied_pattern.observer_patterns) == len(the_pattern.observer_patterns)


def test_change_label(simple_pattern_factory):
    the_pattern = simple_pattern_factory("Pattern1")
    the_pattern.change_label("New label")
    assert the_pattern.label == "New label"
    assert the_pattern.observer_patterns["Observer tag"].label == "New label"


def test_save_and_load_pattern(tmp_path, simple_pattern_factory):
    the_pattern = simple_pattern_factory("Pattern1")
    the_pattern.save(tmp_path, "test")
    loaded_pattern = DummyPattern.load(tmp_path, "test")
    assert the_pattern.label == loaded_pattern.label
    connectors = the_pattern.connectors
    loaded_connectors = loaded_pattern.connectors
    assert len(connectors) == len(loaded_connectors)
    for c, lc in zip(connectors.values(), loaded_connectors.values()):
        assert c.label == lc.label
    observers = the_pattern.observer_patterns
    loaded_observers = loaded_pattern.observer_patterns
    assert len(observers) == len(loaded_observers)
    assert set(observers.keys()) == set(loaded_observers.keys())
